#!/usr/bin/perl
use 5.8.1;
use strict;
#-------------------------------------------------------------------------------
# Relay for HTTP Pseudo-Streaming with Windows Media Encoder
#                                             Copyright (C)2005-2010 nabe@abk
#-------------------------------------------------------------------------------
#[TAB=8][EUC-JP]	Lisenced on GPL(version 2 or later).
#			Sorry, comment is Japanese.
#
# 2010/04/10	Version 2.13	Add -b option
# 2009/08/24	Version 2.12	Windows-GUI complied by ActivePerl5.10.
# 2007/03/08	Version 2.11	Fix not to run on Windows-GUI.
# 2007/02/13	Version 2.10	Windows GUI.
# 2007/02/12	Version 2.02	Fix bug, don't run digest-auth.
# 2007/02/05	Version 2.01	fix on big endian CPU (MacOS etc.)
# 2007/02/03	Version 2.00	Push stream support.
# 2005/06/10	Version 1.01	Client thread change for to sleep.
# 2005/06/03	Version 1.00	First Version
#
our $VERSION = "2.13";
use Socket;
use Fcntl;	# for sysopen
use Digest::MD5 qw(md5_hex);
use threads;
use threads::shared;
#-------------------------------------------------------------------------------
#  - ݒ
#-------------------------------------------------------------------------------
my $bind_port  = 8080;			# default
my $client_max = 100;			# őNCAg
my $log_file   = "wmrelay_%p.log";	# Ot@C

#-----------------------------
# ȉ͒ʏύXsv
#-----------------------------
my $debug;
my $nosilent  = 1;
my $timeout   = 0.2;			# select  Timeout ݒ
my $down_wait = 15;			# I҂i悻 $timeout * $down_waitj
my $bufsize   = 64;			# obt@TCYBonly 2^n
my $realm="wmrelay server";		# DigestF؊֌W
my $WinGUI   = ($^O eq 'MSWin32' || $^O eq 'MSWin64');

my $auth_test;
my $nonce_base = 'KUGX67jPtTekHhsAOZJBgRYobdxScWfm9VQay8IMNLFinElDqrCvupz32w54./10';
my $nonce_size = 16;

#-------------------------------------------------------------------------------
# X^[gAbvƈ
#-------------------------------------------------------------------------------
my $help = 0;
my %s;
my @ary;
while(my $a=shift(@ARGV)) {
	if (ord($a) != 0x2d) { push(@ary, $a); next; }
	my $x = substr($a, 1);
	if    ($x eq "h")  { $help  = 1; }
	elsif ($x eq "d")  { $debug = 1; }
	elsif ($x eq "dd") { $debug = 2; }
	elsif ($x eq "s")  { $nosilent  = 0; }
	elsif ($x eq "t")  { $auth_test = 1; }
	elsif ($x eq "c")  { $WinGUI    = 0; }
	else { $s{$x}=shift(@ARGV); }
}
my $url = shift(@ary);
my $host;
my $port;
my $path;
if ($url =~ m|(?:\w+://)([^/:]*)(?:\:(\d*))?(\/?.*)|) {
	$host = $1;
	$port = $2;
	$path = $3;
} elsif ($url ne '') { $host = $url; }
if ($path eq '') { $path='/'; }
if (defined $ary[0] && $ary[0] =~ /^\d+$/) { $port = int(shift(@ary)); }

if ($help) {
	print <<TEXT;
Usage: $0 [options] [http://URL]
	-h		view this message.
	-s		silent mode
	-p bind_port	bind port number (default $bind_port).
	-m max_conn	max connections (default $client_max).
	-l log_file	log file name (default $log_file).
			replace %p macro with listen port number.
	-f secret_file	push server mode password file.
	-a password	push server mode password (anyuser).
	-b bufsize	ring buffer size 2^n only (max 1024).
	If not set http://URL/, run on push server mode.

	(See more info) http://nabe.blog.abk.nu/0124
TEXT
	exit(0);
}

if (defined $s{p}) { $bind_port  = int($s{p}); }
if (defined $s{m}) { $client_max = int($s{m}); }
if (defined $s{l}) { $log_file   = $s{l};    }
if (defined $s{b} && 0<$s{b}) {
	$bufsize=1024;
	foreach(my $i=16; $i<=1024; $i+=$i) {
		if ($s{b}>$i) { next; }
		$bufsize=$i;
		last;
	}
}
my $bufmask = $bufsize - 1;

#-------------------------------------------------------------------------------
# [{̗pϐ錾
#-------------------------------------------------------------------------------
my $server_down : shared;	# T[oIptO
my $ctrlc_down  : shared;	# CTRL-C ɂT[o_E
my $DATA_header : shared;	# Stream data Header
my @end_thread  : shared;	# IXbh
my $seq      : shared;		# sequence number
my @buffer   : shared;		# NCAgsocket ꎞL^z

my $time     : shared;		# Jn
my $rec_byte : shared;		# Mf[^TCY
my $rec_seq  : shared;		# Mf[^
my @thread_id_pool : shared;	# thread id ێz
#-------------------------------------------------------------------------------
# vÕX^[g
#-------------------------------------------------------------------------------
my @winlog;
my $win;
my $log_fh;
if ($WinGUI) {
	share(@winlog);
	&WinMain();
} else {
	&logmsg("*** Program Start");
	$log_file =~ s/%p/$bind_port/;
	if ($log_file) { sysopen($log_fh, $log_file, O_CREAT | O_WRONLY | O_APPEND); }
	&reley_start();
}
exit(0);

#################################################################################
my $S_listen;
my $HTTP_header="";

sub reley_start {
#################################################################################
#-------------------------------------------------------------------------------
# Ot@C̃I[v
#-------------------------------------------------------------------------------
my $perl_v = $];
$perl_v =~ s/\.?00/./g;
&logmsg("*** Version : $VERSION with Perl $perl_v ($^O)");
&logmsg("*** Log file name : $log_file");
&logmsg("*** Ring buffer size : $bufsize");

#-------------------------------------------------------------------------------
# w|[gListen
#-------------------------------------------------------------------------------
{
	&logmsg("*** Connections max : $client_max");
	&logmsg("*** Bind port : $bind_port");
	socket($S_listen, PF_INET, SOCK_STREAM, 0)
		|| &fail("socket failed: $!") || return 1;
	setsockopt($S_listen, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))
		|| &fail("setsockopt failed: $!") || return 1;
	bind($S_listen, sockaddr_in($bind_port, INADDR_ANY))
		|| &fail("bind port failed: $!") || return 1;
	listen($S_listen, $client_max)
		|| &fail("listen failed: $!") || return 1;
	if ($bind_port == 0) {
		&fail("port number error"); return 1;
	} 
}

#-------------------------------------------------------------------------------
# popT[o[hipɐڑɍsj
#-------------------------------------------------------------------------------
if ($host ne '') {
	my $ip_bin   = inet_aton($host);
	if (!defined $ip_bin) {
		&logmsg("Lookup failed '$host'\n");
		return 1;
	}
	my $ip_adr   = join(".", unpack("C4", $ip_bin));
	my $sock_adr = pack_sockaddr_in($port, $ip_bin);

	&logmsg("*** Connect to http://$host:$port$path");
	my $S_get;
	socket ($S_get, PF_INET, SOCK_STREAM, 0) || &fail("socket failed: $!")     || return 1;
	connect($S_get, $sock_adr)		 || &fail("Connection failed. $!") || return 1;

	# f[^擾NGXg
	&HTTP_REQUEST_get_stream($S_get);
	while(<$S_get>) {
		$HTTP_header .= $_;
		if ($_ eq "\r\n" || $_ eq "\n") { last; }
	}
	&server_main($S_get, $S_listen, 0);
	&logmsg("*** Close receive connection.");
	&logmsg("*** Program shutdown\n");
	close($S_get);
	close($S_listen);
	return 0;
}
#-------------------------------------------------------------------------------
# standalone[hipushp[hj
#-------------------------------------------------------------------------------
&logmsg("*** Standalone deamon mode (push server)");
my $file = $s{f};
my $anypass = $s{a};
my %user;
if ($file ne '') {
	&logmsg("*** Open user password file '$file'");
	sysopen(my $fh, $file, O_RDONLY);
	while(<$fh>) {
		if (ord($_) == 0x23) { next; }	# '#'Ŏn܂s̓Rg
		$_ =~ s/[\r\n]//g;
		my ($user, $pass) = split('=', $_);
		if ($user ne '' && $pass ne '') {
			$user{$user} = $pass;
		}
	}
	close($fh);
}
my $digest;
if ($anypass ne '' || %user) {
	$digest = 1;
	&logmsg("*** Server is digest-authrization mode.");
	if ($anypass ne '') {
		&logmsg("*** Use shared password.");
	}
} else {
	$auth_test = 0;
	&logmsg("*** Users not exists.");
	&logmsg("!!! Server is anonymouse authrization mode.");
}
if ($auth_test) { &logmsg("!!! Authorization test mode."); }
#-----------------------------------------------------------
# p󂯕t[v
#-----------------------------------------------------------
&set_bit(my $listen_bits, $S_listen);
# HTTPwb_̐
$HTTP_header = <<TEXT;
HTTP/1.0 200 OK
Server: Rex/9.0.0.2980
Cache-Control: no-cache
Pragma: no-cache
Pragma: client-id=31415926535
Pragma: features="broadcast,playlist"
Content-Type: application/x-mms-framed

TEXT
my %nonce_buf;
my %authed_push_id;
&logmsg("****************************");
&logmsg("*** Push stream waitting ***");
&logmsg("****************************");
$server_down=0;
while(1) {
	my $r = select(my $x = $listen_bits, undef, undef, 1);
	if ($server_down) { last; }
	if (!$r) { next; }

	# ڑaccept
	my $sock_addr = accept(my $sock, $S_listen);
	if (! $sock_addr) { next; }

	# NCAg̎擾
	my($port, $ip_bin) = sockaddr_in($sock_addr);
	my $ip   = inet_ntoa($ip_bin);
	my $host = gethostbyaddr($ip_bin, AF_INET);
	&logmsg("*** Connection from $host");

	# NGXg̎M
For_keep_alive:
	my $h = &receive_HTTP_header( $sock );
	my $method = $h->{method};
	# NGXg̃`FbN
	if (!$auth_test && ($method ne 'POST'	# POSTł͂Ȃ
	 || ($digest && $h->{'X-ACCEPT-AUTHENTICATION'} !~ /Digest/i) )) {	# Digest F؂łȂ
		&HTTP_RESPONSE_bad_request($sock);
		&logmsg("*** Connection close (Bad Request)");
		close($sock);
		next;
	}
	# push-id ̎擾
	my $push_id;
	if ($h->{COOKIE} =~ /push\-id=(\d+)/) {
		$push_id = $1;
	}
	if ($push_id) { &logmsg("*** Detected push-id=$push_id"); }
	else {
		$push_id = int(rand(2147483640))+7;
		delete $authed_push_id{$push_id};
		&logmsg("*** Generated push-id=$push_id");
	}
	my $auth_ok = $authed_push_id{$push_id};	# cookie session
	# Reject `FbN
	if ($auth_ok < 0) {	# CTRL-CȂǂŕꂽRlNV
		&HTTP_RESPONSE_bad_request($sock);
		&logmsg("*** Connection close (Reject push-id)");
		close($sock);
		delete $authed_push_id{$push_id};	# ͋
		next;
	}
	# DigestF
	if (!$auth_ok && $digest) {
		# F؂̊mF
		while($h->{"AUTHORIZATION"}) {
			&logmsg('*** Client try Digest-authorized');
			my $auth = $h->{"AUTHORIZATION"};
			if (substr($auth,0,7) ne 'Digest ') { last; }
			$auth = substr($auth,7);
			# Authorization wb_̉
			my %auth;
			$auth =~ s/(\w+)=(?:\"([^\"]*)\"|([^\s,]+))/$auth{$1}="$2$3",""/eg;
			# Digestp̏mF
			if ($auth{realm} ne $realm || $auth{qop} ne 'auth'
			 || $auth{algorithm} ne 'MD5-sess') { last; }
			# nonce ̊mFisĂnoncelHj
			my $nonce = $auth{nonce};
			if (!exists $nonce_buf{$nonce}) { last; }
			delete $nonce_buf{$nonce};	# Pxgpnonce̍폜
			# pX[h̃[h
			my $uname=$auth{username};
			my $pass =$anypass;
			if ($pass eq '') {
				if (!exists $user{$uname}) { last; }
				$pass=$user{$uname};
			}
			$auth_test && &logmsg("*** User \"$uname\" with password \"$pass\"");
			# Digest responsel̊mF
			my $cnonce    = $auth{cnonce};
			my $cresponse = $auth{response};
			my $ha1 = md5_hex( md5_hex("$uname:$realm:$pass") . ":$nonce:$cnonce");
			my $ha2 = md5_hex("$method:$auth{uri}");
			my $response = md5_hex("$ha1:$nonce:$auth{nc}:$cnonce:auth:$ha2");
			&logmsg("*** User=\"$uname\" nonce=\"$nonce\" cnonce=\"$cnonce\"");
			&logmsg("*** Client response=\"$cresponse\"");
			&logmsg("*** Server response=\"$response\"");
			if ($cresponse eq $response) {
				$auth_ok=1;
				$authed_push_id{$push_id}=1;
				&logmsg("*** Client Authorized!");
			}
		}
		if (!$auth_ok) {
			my $nonce = &generate_nonce();
			$nonce_buf{$nonce} = $ip;
			&logmsg("*** 401_unauthorized (nonce \"$nonce\")");
			&HTTP_RESPONSE_401_unauthorized($sock, $push_id, $nonce);
			close($sock);
			next;
		}
	}
	### 󂯓ok
	&logmsg("*** Responsed push-ok to $host");
	if ($h->{'CONTENT-LENGTH'} == 0) {
		# ConnectionmF Content-Length = 0
		&HTTP_RESPONSE_pushsetup( $sock, $push_id );
		if ($h->{CONNECTION} eq 'Keep-Alive') {
			&logmsg("*** Keep alive connection [$host]");
			goto For_keep_alive;
		}
	} else {
		# pT[oN
		my $sig_int = $SIG{INT};
		&server_main($sock, $S_listen, 1);
		$SIG{INT} = $sig_int;
		&logmsg("!!! push-id=$push_id is set Reject flag.");
		$authed_push_id{$push_id}=-1;
	}
	# StreamM
	&logmsg("*** Close receive connection.");
	close($sock);

	if ($WinGUI && $ctrlc_down) { last; }
	# ̃[vp
	$server_down=0;
	&logmsg("****************************");
	&logmsg("*** Push stream waitting ***");
	&logmsg("****************************");
}
close($S_listen);
&logmsg("*** Server Down\n");
return 0;
#################################################################################
}

#################################################################################
# T[ovZX
#################################################################################
my $push_mode;
my $receive_bits;
sub server_main {
#-------------------------------------------------------------------------------
# T[o[C
#-------------------------------------------------------------------------------
	my ($S_receive, $S_listen, $pm) = @_;
	$push_mode = $pm;
	#--------------------------------------------------
	# nh̐ݒ
	#--------------------------------------------------
	$server_down = 0;
	$ctrlc_down  = 0;
	local($SIG{INT})  = \&server_down;
	local($SIG{PIPE}) = sub { return ; };

	#--------------------------------------------------
	# \Pbg֘A̕ϐݒ
	#--------------------------------------------------
	my $select_bits  ='';
	my $listen_bits  ='';
	   $receive_bits ='';
	&set_bit($select_bits,  $S_listen);
	&set_bit($listen_bits,  $S_listen);
	&set_bit($select_bits,  $S_receive);
	&set_bit($receive_bits, $S_receive);
	my $s_receive_fileno = fileno($S_receive);
	my $s_listen_fileno  = fileno($S_listen);

	#--------------------------------------------------
	# [vȌ
	#--------------------------------------------------
	$time = time;		# Jn
	$rec_byte = 0;		# Mf[^TCY
	$rec_seq = 0;		# Mf[^
	@thread_id_pool = ();	# thread id ێz
	my @threads;		# threads object ۑp
	for(my $i=0; $i<$client_max; $i++) {
		$thread_id_pool[$i] = $client_max - $i;
	}
	#--------------------------------------------------
	# C[v
	#--------------------------------------------------
	&logmsg("*******************");
	&logmsg("*** Relay start ***");
	&logmsg("*******************");
	while(1) {
		# ڑĎ
		if ($server_down) { $select_bits = $listen_bits; }
		my $r = select(my $x = $select_bits, undef, undef, $timeout);
		if ($r < 0) { $x=''; }	# CTRL-C Ȃ
		#----------------------------------------------------
		# VȐڑ
		#----------------------------------------------------
		if (vec($x, $s_listen_fileno, 1) ) {
			# Xbh𐶐ڑ Accept 
			my $sock_addr = accept(my $S_client, $S_listen);
			my $thread_id = pop(@thread_id_pool);
			$debug && logmsg("[thread:$thread_id] create");
			$threads[ $thread_id ] = threads->create(\&accept_listener, $S_client, $sock_addr, $thread_id);
			close($S_client);
		#----------------------------------------------------
		# pf[^M
		#----------------------------------------------------
		} elsif (!$server_down && vec($x, $s_receive_fileno, 1) ) {
			# $debug && &logmsg("new data from stream server");
			my $size = &read_stream_1block( $S_receive, my $data );
			if (!$size) {
				&logmsg("*** relay-server down (please wait).");
				$server_down=1; next;
			}
			$rec_byte += $size;
			$rec_seq++;
		#----------------------------------------------------
		# NCAgڑ`FbN
		#----------------------------------------------------
		} else {
			# IXbh join  thread_id ėp\ɂ
			if (@end_thread) {
				my @ary = @end_thread; @end_thread = ();
				foreach(@ary) {
					my $obj = $threads[$_];
					if ($obj) {
						$obj->join();
						$threads[$_] = undef;
						if ($_) { push(@thread_id_pool, $_); }
					}
				}
				&display_connection_num();
			}
			# Server down check
			if ($server_down) {
				if (++$server_down > $down_wait) { last; }
			}
		}
	}
	#--------------------------------------------------
	# C[vI
	#--------------------------------------------------
	$time     = format_time( time - $time );
	$rec_byte = format_byte( $rec_byte );
	&logmsg("*** time $time, $rec_byte byte received. $rec_seq sequences.");

	# NCAgXbhI
	foreach(@threads) { if (defined $_) { $_->join(); } }

	# [T[õ[vI
	return $ctrlc_down;
}

#-------------------------------------------------------------------------------
# CTRL-C̏
#-------------------------------------------------------------------------------
sub server_down {	# CTRL-C
	print "\n";
	if ($server_down) { $server_down=9999999; return ; }
	$server_down=2;
	$ctrlc_down =1;
	&send_eos();
	&logmsg("*** relay-server down (please wait).");
}
#-------------------------------------------------------------------------------
# RlNV̕\
#-------------------------------------------------------------------------------
sub display_connection_num {
	my $str = sprintf("[%d / $client_max]", $client_max - $#thread_id_pool -1);
	&logmsg("*** connections $str");
}
#-------------------------------------------------------------------------------
# Xi[̐ڑ
#-------------------------------------------------------------------------------
my $client_loop;
my $client_id;
sub accept_listener {
	my ($S_client, $sock_addr, $thread_id) = @_;
	$win=undef;
	# RlNV\
	&display_connection_num();
	# C
	my $r = &accept_listener_main( @_ );
	# I
	close($S_client);
	push(@end_thread, $thread_id);
	# fobOO
	$debug && logmsg("[thread:$thread_id] end");
	return $r;
}
sub accept_listener_main {
	my ($sock, $sock_addr, $thread_id) = @_;
	# accept G[
	if (! $sock_addr) { close($sock); return -1; }
	# NCAg̎擾
	my($port, $ip_bin) = sockaddr_in($sock_addr);
	my $ip   = inet_ntoa($ip_bin);
	my $host = gethostbyaddr($ip_bin, AF_INET);
	my $id   = "$ip:$port";		# NCAgID
	&logmsg("[$id] Connection from $host");

	# őڑɂĂȂmF
	if (! $thread_id) {
		&logmsg("[$id] $client_max connected max. connection close.");
		&HTTP_RESPONSE_client_max( $sock );
		return -2;
	}

	# HTTP wb_M
	my $h = &receive_HTTP_header( $sock );
	my $method = $h->{method};

	# Windows media steram Ȃ
	# asx t@C̎擾H
	if ($method eq 'GET' && $h->{HOST} && $h->{path} =~ /\.asx$/) {
		&logmsg("[$id] client needed asx file ($h->{path})");
		my $title = $s{t} || 'wmrelay stream';
		$title =~ s/[\x00-\x1f<>\"\']//g;
		my $asx = <<TEXT;
<ASX version="3.0">
	<Entry>
		<Title>$title</Title>
		<Ref href="http://$h->{HOST}/stream.asf"/>
	</Entry>
</ASX>
TEXT
		&HTTP_RESPONSE_data( $sock, $asx, 'video/x-ms-asx');
		return 3;
	}
	if (!$method || $h->{'USER-AGENT'} !~ /^NSPlayer/) {
		&HTTP_RESPONSE_bad_request($sock);
		&logmsg("[$id] close (Bad Request, client get mp3 stream)");
		return 1;
	}
	if ($method eq 'POST' && $h->{log_line}) {		# log POST
		&logmsg("[$id] log-line from client : $h->{log_line}");
		return 2;
	}

	#---------------------------------------------------
	# Xg[~Of[^̃wb_]
	#---------------------------------------------------
	# HTTP wb_o
	socket_out($sock, $HTTP_header);
	# f[^ubNwb_o
	socket_out($sock, $DATA_header);

	$client_loop=1;
	$client_id  =$id;
	local($SIG{PIPE}) = \&broken_client_pipe;
	my $start_time = time;	# Jn
	my $byte  = 0;		# MoCg
	my $myseq = $seq;	# ݂̃V[PXԍ
	while($client_loop && !$server_down) {
		select(my $x = $receive_bits, undef, undef, $timeout);
		if ($seq <= $myseq) { next; }

		# VȎMf[^NCAgɑM
		while($myseq < $seq) {
			$myseq++;
			my $bufnum = $myseq & $bufmask;
			my $buflen = length($buffer[$bufnum]);
			$debug && &logmsg("[$id] send data $buflen byte, seq=$myseq");
			my $r = syswrite($sock, $buffer[$bufnum], $buflen);
			$byte += $buflen;
			# lbg[NxȂǂŔтȂΓrf[^j
			if ($seq - $bufsize > $myseq) { $myseq=$seq; last; }
			if (!defined $r) { $client_loop=0; last; }	# ݎsbroken pipe
		}
	}
	my $time = &format_time( time - $start_time );
	   $byte = &format_byte( $byte );
	&logmsg("[$id] close. time $time, $byte sended.");
	return 0;
}

sub broken_client_pipe {
	$client_loop=0;
	&logmsg("[$client_id] pipe breoken");
};

#-------------------------------------------------------------------------------
# Xg[PubNM
#-------------------------------------------------------------------------------
sub read_stream_1block {
	my $fh = shift;
	# Read 1 block
	read($fh, my $head,     2);
	read($fh, my $size_bin, 2);

	if (ord($head) != 0x24 && $head ne '?D') {		# unknown block
		$head =~ s/([\x00-\x1f\x7f-\xff])/'[' . unpack('H2',$1) . ']'/eg;
		&logmsg("*** Receive data faild (unknown block '$head')");
		return 0;
	}
	# Save block header

	# Read block data
	my $data;
	my ($size, $sequence);
	if ($push_mode) {
		# vbVT[o[h
		$size = unpack('v', $size_bin);
		$size_bin = pack('v', $size+8);	# MSS Pre-HeaderTCY̒ǉ
		my $id_flags;
		if ($head eq '$H') {
			$sequence = 0;
			$id_flags = "\x00\x0c";
		} else {
			$sequence = $seq+1;
			$id_flags = "\x00\x00";
		}
		# MSS Pre-header ǉ
		$data = $head . $size_bin . pack('V', $sequence) . $id_flags . $size_bin;
		read($fh, $data, $size, 12);
	} else {
		# ʏ pop [[h
		$data = $head . $size_bin;
		$size = unpack('v', $size_bin);
		read($fh, $data, $size, 4);
		$sequence = unpack('V', substr($data,4,4));
	}
	# End of stream
	if ($head eq '$E') {
		$nosilent && &logmsg("Recived stream $head : $size byte, seq=$sequence [End of Stream]");
		&send_eos();
		return 0;
	}
	# bZ[W\
	$nosilent && &logmsg("Recived stream $head : $size byte, seq=$sequence");

	# save to ring buffer
	$buffer[$sequence & $bufmask] = $data;
	$seq = $sequence;

	# DATA Header save
	if ($head eq '$H') {
		$DATA_header = $data;
	}
	return $size + 4;
}

#-------------------------------------------------------------------------------
# Xg[PubNM
#-------------------------------------------------------------------------------
sub send_eos {
	my $newseq = $seq+1;
	my $eos    = '$E' . pack('v', 4) . pack('V', 1);
	$debug && &logmsg("*** EOS save to buffer seq=$newseq");
	$buffer[$newseq & $bufmask] = $eos;
	$seq = $newseq;
}

#################################################################################
# HTTP/socket Tu[`
#################################################################################
#-------------------------------------------------------------------------------
# HTTPwb_̓[`
#-------------------------------------------------------------------------------
sub receive_HTTP_header {
	my $fh = shift;
	my %h;
	my $str = <$fh>;
	$str =~ s/\r\n/\n/;
	if ($str =~ /^(GET|POST) ([^\s]+) HTTP\/(\d\.\d)/) {
		$h{method} = $1;
		$h{path}   = $2;
		$h{http}   = $3;
	} elsif ($str =~ m|^HTTP/(\d.\d) (\d+) (\w+)|) {
		$h{http}       = $1;
		$h{status}     = $2;
		$h{status_msg} = $3;
	} else {
		$debug && &logmsg("[HTTP] Bad HTTP request or response");
		$h{error} = 1;
		return \%h;	# bad request
	}

	while(<$fh>) {
		$_ =~ s/\r\n/\n/;
		if ($_ eq "\n") { last; }
		$str .= $_;
		chomp($_);
		if (substr($_, 0, 17) eq "Pragma: log-line=") {
			$h{log_line} = substr($_, 17);
			next;
		}
		if ($_ =~ /^([\w\-]+): (.+)/) {
			my $key = uc($1);
			while (exists $h{$key}) { $key .= '_'; }
			$h{$key} = $2;
		}
	}
	if (0 && $debug) {
		foreach(sort(keys(%h))) {
			print "$_=$h{$_}\n";
		}
	}
	$h{string} = $str;
	($debug >1) && print $str;
	return \%h;
}


#-------------------------------------------------------------------------------
# HTTPwb_[`
#-------------------------------------------------------------------------------
sub HTTP_REQUEST_get_stream {
	my $fh = shift;
	my $p="";
	if ($port != 80) { $p=":$port"; } 
	&socket_out($fh, <<TEXT);
GET $path HTTP/1.0
Accept: */*
User-Agent: NSPlayer/4.1.0.3928
Host: $host$p
Pragma: no-cache
Content-Type: application/x-mms-framed

TEXT
}
#----------------------------------------------------------
sub HTTP_RESPONSE_bad_request {
	my $fh = shift;
	&socket_out($fh, <<TEXT);
HTTP/1.0 400 Bad Request
Server: Rex/10.0.0.3650.1
Cache-Control: no-cache
Pragma: no-cache
Pragma: client-id=1337885143

TEXT
}
#----------------------------------------------------------
sub HTTP_RESPONSE_client_max {
	my $fh = shift;
	&socket_out($fh, <<TEXT);
GET / HTTP/1.0
HTTP/1.0 501 Not Implemented
Server: Rex/9.0.0.2980.1
Cache-Control: no-cache
Pragma: no-cache

TEXT
}
#----------------------------------------------------------
sub HTTP_RESPONSE_data {
	my $fh      = shift;
	my $content = shift;
	my $type    = shift || 'text/plain';
	my $length  = length($content);
	&socket_out($fh, <<TEXT);
HTTP/1.0 200 OK
Server: wmrelay.pl
Content-Length: $length
Content-type: $type
Cache-Control: no-cache
Pragma: no-cache

$content
TEXT
}

#----------------------------------------------------------
sub HTTP_RESPONSE_pushsetup {
	my $fh   = shift;
	my $pid  = shift;
	my $date = &get_rfc_date();
	&socket_out($fh, <<TEXT);
HTTP/1.1 204 No Content
Server: Cougar/9.01.01.3814
Content-Length: 0
Date: $date
Pragma: no-cache, timeout=60000
Cache-Control: no-cache
Set-Cookie: push-id=$pid

TEXT
}

#----------------------------------------------------------
sub HTTP_RESPONSE_401_unauthorized {
	my $fh    = shift;
	my $pid  = shift;
	my $nonce = shift;
	my $date  = &get_rfc_date();
	my $content = <<TEXT;
<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
<html><head>
<title>401 Authorization Required</title>
</head><body>
<h1>Authorization Required</h1>
<p>This windows media stream server is needed digest authorize.</p>
<p>This page is generated by "wmrelay.pl".</p>
</body></html>
TEXT
	my $length = length($content);
	&socket_out($fh, <<TEXT);
HTTP/1.1 401 Unauthorized
Server: Cougar/9.01.01.3814
WWW-Authenticate: Digest qop="auth",algorithm=MD5-sess,nonce="$nonce",realm="$realm"
Date: $date
Pragma: no-cache, timeout=60000
Set-Cookie: push-id=$pid
Content-Length: $length
Content-type: text/html
Supported: com.microsoft.wm.srvppair, com.microsoft.wm.sswitch, com.microsoft.wm.predstrm, com.microsoft.wm.fastcache, com.microsoft.wm.startupprofile

$content
TEXT
}

#-------------------------------------------------------------------------------
# Tu[`
#-------------------------------------------------------------------------------
sub set_bit	{ vec($_[0], fileno($_[1]), 1) = 1; }
sub reset_bit	{ vec($_[0], fileno($_[1]), 1) = 0; }
sub fail {
	&logmsg(@_);
	if (! $WinGUI) { exit(1); }
	return 0;
}
sub socket_out {
	my $socket = shift;
	syswrite($socket, $_[0], length($_[0]) );
}
sub generate_nonce {
	my $size = shift;
	my $nonce='';
	foreach(1..$nonce_size) {
		$nonce .= substr($nonce_base, int(rand(64)), 1);
	}
	return $nonce;
}

sub get_rfc_date {
        my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime(time);
        my($wd, $mn);
        $wd = substr('SunMonTueWedThuFriSat',$wday*3,3);
        $mn = substr('JanFebMarAprMayJunJulAugSepOctNovDec',$mon*3,3);
        return sprintf("$wd, %02d $mn %04d %02d:%02d:%02d GMT"
                , $mday, $year+1900, $hour, $min, $sec);
}

sub format_time	{
	my ($sec) = @_;
	my $min  = int($sec/60);
	$sec -= $min  * 60;
	my $hour = int($min/60);
	$min -= $hour * 60;
	return sprintf("%02d:%02d:%02d", $hour, $min, $sec);
}
sub format_byte	{
	my ($byte) = @_;
	my $z = "Byte";
	if ($byte > 1024) { $byte >>= 10; $z = "KB"; }
	if ($byte > 1024) { $byte >>= 10; $z = "MB"; }
	return "$byte$z";
}

sub logmsg {
	my $tm  = &get_timestamp();
	my $msg = "$tm @_\n";
	if ($WinGUI) { push(@winlog, $msg); return; }
	if ($log_fh) { print $log_fh $msg; }
	print $msg;

}
sub get_timestamp {
	my ($sec,$min,$hour, $day,$mon,$year, $wday) = localtime();
	$year += 1900; $mon++;
	$mon  = sprintf('%02d', $mon);
	$day  = sprintf('%02d', $day);
	$hour = sprintf('%02d', $hour);
	$min  = sprintf('%02d', $min);
	$sec  = sprintf('%02d', $sec);

	return "$year/$mon/$day $hour:$min:$sec";
}

###############################################################################
###############################################################################
###############################################################################


###############################################################################
# WindowspGUI[`
###############################################################################
#------------------------------------------------------------------------------
# Windows main
#------------------------------------------------------------------------------
my $relay_running = undef;
my $timer;
sub WinMain {
	require Win32::GUI;
	require Win32::GUI::Loft::Design;
	my $filename = "wmrelay.gld";
	my $gld = read_file( $filename );
	my $design = Win32::GUI::Loft::Design->newScalar($gld, $filename) or die("Fail newScalar");
	$win = $design->buildWindow() or die("Could not build window");
	$win->Text("wmrelay $VERSION (C)nabe\@abk");

	$win->Show();
	$win->SetForegroundWindow();
	$win->SetActiveWindow();

	# _CAOl
	$win->tfPortNumber->Text( $bind_port );
	$win->tfMaxConnection->Text( $client_max );
	$win->tfLogFile->Text( $log_file );

	# ^C}[o^
	$timer = $win->AddTimer( "system", 100 );	# 0.1b

	# X^[gAbv
	share( $relay_running );
	$nosilent = 0;
	&logmsg("*** Program Start");
	Win32::GUI::Dialog();
}

#------------------------------------------------------------------------------
# WindowsCxg
#------------------------------------------------------------------------------
my $rec;
sub ::system_Timer {
	# O̓]
	while(@winlog) {
		&win_logmsg(shift(@winlog));
	}
	# [T[o쒆
	if ($relay_running && $rec<$rec_byte) {
		my $byte  = &format_byte( $rec_byte );
		my $times = &format_time( time - $time );
		my $conn  = $client_max - $#thread_id_pool -1;
		$win->tfRecBytes->Text( $byte  );
		$win->tfRecTimes->Text( $times );
		$win->tfConnections->Text( $conn );
	# [T[o̒~
	} elsif ($relay_running eq '0') {
		$relay_running = undef;
		$win->btnMain->Text('X^[g');
		close($log_fh);
	}
	$rec = $rec_byte;
}

my $server_thread;
sub ::btnMain_Click {
	if ($relay_running) {
		&logmsg('*** Stopping');
		$server_down = 1;	# GoTo Down
		$ctrlc_down  = 1;
	} else {
		if ($server_thread) {
		#	$server_thread->join();
			$server_thread = undef;
		}
		&win_server_start();
	}
}

sub win_server_start {
	$win->btnMain->Text('Xgbv');
	&logmsg('*** Starting server process');
	# ݒl̎荞
	$bind_port  = int( $win->tfPortNumber->Text() );
	$client_max = int( $win->tfMaxConnection->Text() );
	$log_file   = $win->tfLogFile->Text();

	$s{f} = $s{a} = $url = $host = $port = $path = undef;
	if ( $win->rbRecSelect0->Checked() ) {	# GET [h
		$host = 'unknown';
		$url = $win->tfURL->Text();
		if ($url =~ m|(?:\w+://)([^/:]*)(?:\:(\d*))?(\/?.*)|) {
			$host = $1 || 'unknown';
			$port = $2;
			$path = $3;
		}
	} elsif ( $win->rbRecSelect1->Checked() ) {	# LpX[h[h
		$s{a} = $win->tfPass->Text();
	} elsif ( $win->rbRecSelect1->Checked() ) {	# [U[ʃpX[ht@C
		$s{f} = $win->tfPassFile->Text();
	}

	# 
	$win->tfRecBytes->Text( '0' );
	$win->tfRecTimes->Text( '00:00:00' );
	$win->tfConnections->Text( '0' );

	# Ot@C̃I[v
	$log_file =~ s/%p/$bind_port/;
	if ($log_file) { sysopen($log_fh, $log_file, O_CREAT | O_WRONLY | O_APPEND); }

	$timer->Kill(1);
	$timer = undef;
	$server_thread = threads->create(\&win_relay_start);
	$timer = $win->AddTimer( "system", 100 );	# 0.1b
	$relay_running = 1;
}

# T[oXbhĂяoi]vȕϐRs[Ȃj
sub win_relay_start {
	close($log_fh);
	$log_fh=undef;
	$win=undef;
	$server_thread=undef;
	$rec=$rec_byte=0;
	&reley_start();
	$relay_running = 0;
}


#------------------------------------------------------------------------------
# Windowp[`
#------------------------------------------------------------------------------
# Ot@CI{^
sub ::btnLogFile_Click {
	my $ret = Win32::GUI::GetSaveFileName(
		-owner		=> $win,
		-title		=> 'Ot@CI',
		-file		=> "wmrelay_%p.log",
		-filter		=> ['Ot@C', '*.log', 'ׂẴt@C', '*.*'],
		-directory	=> '.\\'
	);
	$win->tfLogFile->Text($ret);
}

# pX[ht@CI{^
sub ::btnPassFile_Click {
	my $ret = Win32::GUI::GetOpenFileName(
		-owner		=> $win,
		-title		=> '[U[pX[ht@CI',
		-file		=> "wmrelay_users.txt",
		-filter		=> ['ׂẴt@C', '*.*'],
		-directory	=> '.\\'
	);
	$win->tfPassFile->Text($ret);
}

#------------------------------------------------------------------------------
# WindowpTu[`
#------------------------------------------------------------------------------
# for PAR
sub read_file {
	my $file = shift;
	if ($INC{'PAR.pm'}) {	# PAR
		return PAR::read_file($file);
	}
	sysopen(my $fh, $file, O_RDONLY);
	my @ary = <$fh>;
	close($fh);
	return join('', @ary);
}

sub win_logmsg {
	my $msg   = shift;
	my $reMsg = $win->reMsg;
	if (!$reMsg) { return ; }
	my $text  = $reMsg->Text();
	my $len   = length($text);
	$reMsg->SetFocus();
	$reMsg->Select($len, $len);
	$reMsg->ReplaceSel($msg);

	if ($log_fh) { print $log_fh $msg; }
}

